/*
 *            longhaul cpu usage governor
 *
 *  v 1.2
 *
 *  (C) 2003 Andrew Howlett andrew@howlett.net
 *  Licensed under the terms of the GNU GPL License version 2.
 *
 *  This module takes three parameters: refresh_period, up_trigger,
 *  and down_trigger. The module checks cpu usage every refresh_period
 *  seconds. If cpu usage exceeds up_trigger percent, then the module
 *  changes the cpu clock multiplier to the maximum allowed (i.e. the
 *  factory default as indicated in MSR_EBL_CR_POWERON. If the cpu usage
 *  is less than down_trigger, then the module decreases the clock multiplier
 *  by .5 (unless the value is 3, which is the minimum value).
 *
 *    5 <= refresh_period <= 60   default = 10 seconds
 *    5 <= up_trigger     <= 95   default = 33 %
 *    5 <= down_trigger   <= 95   default = 66 %
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>      /* for jiffies and init_task_union */
#include <linux/kernel.h>
//#include <linux/interrupt.h> /* tasklets */
#include <linux/init.h>

#include "longhaul.h"

#define PFX "longhaul cpu usage governor: "

//#define DEBUG
#ifdef DEBUG
#  define dprintk(msg...) printk(msg)
#else
#  define dprintk(msg...) do { } while(0)
#endif /* DEBUG */

/*
 *
 * module parameters
 *
 */

static int refresh_period=0;
static int up_trigger=0;
static int down_trigger=0;
static int log=1;

/*
 * initialize the timer structure, used to time the delay between cpu_usage calculations
 */

struct timer_list lh_governor_timer;

/*
 * Keep track of info we need between timer runs.
 */

static struct lh_governor_data_t {
	int refresh_period;
	int up_trigger;
	int down_trigger;
	unsigned long old_time, old_idle;
} lh_governor_data;

/*
 *  change cpu clocks
 *
 *   increase = 1   then increase the clock to max
 *            = 0   decrease the cpu clock by one level
 */

static void lh_governor_change_cpu_clock(int increase)
{
	int i, highest_clock, current_clock;
	struct longhaul_state_t longhaul_state;

	load_longhaul_state(&longhaul_state);
	
	dprintk(KERN_INFO PFX " change clock %d\n", increase);

	/* the clock table is in order from lowest to highest clock,
	 * so this will set i to the highest clock
	 */
	
	for (i=0; longhaul_state.clock_table[i].frequency != CLOCK_TABLE_END; i++)
		dprintk (KERN_INFO PFX "clock %#x  %06d\n", i, longhaul_state.clock_table[i].frequency);	
	highest_clock = i-1;

	/*     find current clock  */

	for (i=0; longhaul_state.clock_table[i].frequency != longhaul_state.cpu_speed; i++);
	current_clock = i;

	/* note - if overclock is enabled do not change the cpu clock
	 *        governor should not be used with overclock-enabled longhaul driver
	 */
	
	if (!longhaul_state.overclock_on && increase && current_clock < highest_clock) {
		longhaul_set_state(&longhaul_state, highest_clock);
		if (log) printk(KERN_INFO PFX " new clock: %d\n", highest_clock);
	}
	
	if (!longhaul_state.overclock_on && !increase && current_clock>0) {
		longhaul_set_state(&longhaul_state, --current_clock);
		if (log) printk(KERN_INFO PFX " new clock: %d\n", current_clock);
	}
}


/*
 * timer callback
 *
 * really, this should be done with tasklets
 *
 */

static void lh_governor_timeout(unsigned long ptr)
{
 	unsigned long new_time, new_idle;
	unsigned long idle, usage;
	struct lh_governor_data_t *p = (struct lh_governor_data_t *) ptr;

	new_time = jiffies;
	new_idle = init_task.times.tms_utime + init_task.times.tms_stime;
	idle = (new_idle - p->old_idle)*100;
	idle /= (new_time - p->old_time);
	usage = 100 - idle;
	dprintk(KERN_INFO PFX " 100 - ( %lu - %lu ) * 100/(%lu - %lu) = %ld\n", new_idle, p->old_idle,
					new_time, p->old_time, usage);
	if (usage > p->up_trigger) {
		lh_governor_change_cpu_clock(1);
	}
	if (usage < p->down_trigger) {
		lh_governor_change_cpu_clock(0);
	}

	p->old_time = new_time;
	p->old_idle = new_idle;
	lh_governor_timer.function = lh_governor_timeout;
	lh_governor_timer.data = ptr;
	lh_governor_timer.expires = jiffies + p->refresh_period * HZ;
	add_timer(&lh_governor_timer);
}

/*
 *
 *     init_module:  When the module is installed it executes this function
 *
 */

int init_module(void) 
{
	if (refresh_period == 0) refresh_period = 10; /* default value every 10 seconds*/
	if (refresh_period < 1) refresh_period = 1;
	if (refresh_period > 60) refresh_period = 60;

	if (up_trigger == 0) up_trigger = 33;
	if (up_trigger < 5) up_trigger = 5;
	if (up_trigger > 95) up_trigger = 95;
	if (down_trigger == 0) down_trigger = 66;
	if (down_trigger < 5) down_trigger = 5;
	if (down_trigger > 95) down_trigger = 95;

	printk(KERN_INFO PFX "initialized\n");
	dprintk(KERN_INFO PFX "refresh_period: %d  up_trigger: %d  down_trigger: %d\n",
					refresh_period, up_trigger, down_trigger);

	lh_governor_data.up_trigger = up_trigger;
	lh_governor_data.down_trigger = down_trigger;
	lh_governor_data.refresh_period = refresh_period;
	
	/* Push the data into the lh_governor_data structure, then create the timer
	 *
	 */
	
	lh_governor_data.old_time = jiffies;
	lh_governor_data.old_idle = init_task.times.tms_utime + init_task.times.tms_stime;
	lh_governor_timer.function = lh_governor_timeout;
	lh_governor_timer.data = (unsigned long) &lh_governor_data;
	lh_governor_timer.expires = jiffies + lh_governor_data.refresh_period * HZ;
	init_timer(&lh_governor_timer);
	add_timer(&lh_governor_timer);              /* create the timer             */

	return 0;  /* succeed */
}

void cleanup_module(void)
{
	del_timer_sync(&lh_governor_timer);  /* kill timer, must kill timer before tasklet */
	printk(KERN_INFO PFX "removed\n");
}

MODULE_AUTHOR ("A. Howlett <andrew@howlett.net> \n");
MODULE_DESCRIPTION ("VIA C3 Longhaul cpu usage governor.");
MODULE_LICENSE ("GPL");
MODULE_PARM (refresh_period, "i");
MODULE_PARM (up_trigger, "i");
MODULE_PARM (down_trigger, "i");
MODULE_PARM (log, "i");
