/*
 *            Andrew's longhaul module
 *  version 1.2
 *
 *
 *  (C) 2003 Andrew Howlett andrew@howlett.net
 *  Licensed under the terms of the GNU GPL License version 2.
 *
 *  Based on the cpufreq by Dave Jones and Padraig Brady
 *
 *  BIG FAT DISCLAIMER: Work in progress code. Possibly *dangerous*
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <asm/msr.h>
#include "longhaul.h"

//#define DEBUG
#define  OVERCLOCK   0    /* 0 = no overclocking, 1 = allow overclocking */
#define PFX " longhaul "
#define DRIVER_MAJOR_VERSION 1
#define DRIVER_MINOR_VERSION 2

#define __hlt()     __asm__ __volatile__("hlt": : :"memory")

/*
 *
 * stuff not defined in old kernel headers
 *
 */

#ifdef DEBUG
#  define dprintk(msg...) printk(msg)
#else
#  define dprintk(msg...) do { } while(0)
#endif /* DEBUG */

#ifndef MSR_VIA_LONGHAUL
#  define MSR_VIA_LONGHAUL   0x110a
#endif   /* MSR_VIA_LONGHAUL */


/*
 *
 * module parameters
 *
 */

/*
 *
 * globals
 *
 */

static struct longhaul_state_t longhaul_state_buffer;

/* There are three versions of longhaul. Each version of longhaul uses a different
 * table of clock ratios. There are types of VIA CPUs - each type is defined by a
 * triplet of numbers: CPU_FAMILY.CPU_MODEL.STEPPING. The table below maps CPU types
 * to longhaul version.
 *
 *    Family    Model   Stepping     Core        Longhaul
 *      6         6        all      Samuel          1
 *      6         7         0       Samuel 2        1
 *      6         7        1-7      Samuel 2        2
 *      6         7        8-15      Ezra           2
 *      6         8         x        Ezra T         3
 *      6         9         0-1     Nehemiah        4a
 *      6         9         2->     Nehemiah        4b
 *
 */


/* clock tables:   clock ratios * 10
 *                 -1 means reserved
 *                 -2 is end of clocks flag
 */

static int __initdata longhaul2_cpu_mult_table[17] = {
	100, /* 0000 -> 10.0x */
	30,  /* 0001 ->  3.0x */
	40,  /* 0010 ->  4.0x */
	90,  /* 0011 ->  9.0x */
	95,  /* 0100 ->  9.5x */
	35,  /* 0101 ->  3.5x */
	45,  /* 0110 ->  4.5x */
	55,  /* 0111 ->  5.5x */
	60,  /* 1000 ->  6.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	50,  /* 1011 ->  5.0x */
	65,  /* 1100 ->  6.5x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	120, /* 1111 -> 12.0x */
  -2
};

static int __initdata longhaul3_cpu_mult_table[33] = {
	100, /* 0000 -> 10.0x */
	30,  /* 0001 ->  3.0x */
	40,  /* 0010 ->  4.0x */
	90,  /* 0011 ->  9.0x */
	95,  /* 0100 ->  9.5x */
	35,  /* 0101 ->  3.5x */
	45,  /* 0110 ->  4.5x */
	55,  /* 0111 ->  5.5x */
	60,  /* 1000 ->  6.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	50,  /* 1011 ->  5.0x */
	65,  /* 1100 ->  6.5x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	120, /* 1111 ->  12.0x */
	-1,  /* 0000 -> RESERVED (10.0x) */
	110, /* 0001 -> 11.0x */
	120, /* 0010 -> 12.0x */
	-1,  /* 0011 -> RESERVED (9.0x)*/
	105, /* 0100 -> 10.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	135, /* 0111 -> 13.5x */
	140, /* 1000 -> 14.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	130, /* 1011 -> 13.0x */
	145, /* 1100 -> 14.5x */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	-1,  /* 1111 -> RESERVED (12.0x) */
  -2,
};

static int __initdata longhaul4a_cpu_mult_table[33] = {
	100, /* 0000 -> 10.0x */
	160, /* 0001 -> 16.0x */
	-1,  /* 0010 ->  RESERVED */
	90,  /* 0011 ->  9.0x */
	95,  /* 0100 ->  9.5x */
	-1,  /* 0101 ->  RESERVED */
	-1,  /* 0110 ->  RESERVED */
	55,  /* 0111 ->  5.5x */
	60,  /* 1000 ->  6.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	50,  /* 1011 ->  5.0x */
	65,  /* 1100 ->  6.5x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	120, /* 1111 -> 12.0x */
	100, /* 0000 -> 10.0x */
	-1,  /* 0001 -> RESERVED */
	120, /* 0010 -> 12.0x */
	90,  /* 0011 ->  9.0x */
	105, /* 0100 -> 10.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	135, /* 0111 -> 13.5x */
	140, /* 1000 -> 14.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	130, /* 1011 -> 13.0x */
	145, /* 1100 -> 14.5x */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	120, /* 1111 -> 12.0x */
  -2,
};

static int __initdata longhaul4b_cpu_mult_table[33] = {
	100, /* 0000 -> 10.0x */
	160, /* 0001 -> 16.0x */
	-1,  /* 0010 ->  RESERVED */
	90,  /* 0011 ->  9.0x */
	95,  /* 0100 ->  9.5x */
	-1,  /* 0101 ->  RESERVED */
	-1,  /* 0110 ->  RESERVED */
	55,  /* 0111 ->  5.5x */
	60,  /* 1000 ->  6.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	50,  /* 1011 ->  5.0x */
	65,  /* 1100 ->  6.5x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	120, /* 1111 -> 12.0x */
	100, /* 0000 -> 10.0x */
	110, /* 0001 -> 11.0x */
	120, /* 0010 -> 12.0x */
	90,  /* 0011 ->  9.0x */
	105, /* 0100 -> 10.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	135, /* 0111 -> 13.5x */
	140, /* 1000 -> 14.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	130, /* 1011 -> 13.0x */
	145, /* 1100 -> 14.5x */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	120, /* 1111 -> 12.0x */
  -2,
};



/*
 * When the CPU powers up it needs to know what default FSB and CPU multiplier to use.
 * These settings are stored in a machine specific register - specifically the
 * EBL_CR_POWERON MSR. Each type of CPU core has a different
 * EBL_CR_POWERON register. This register contains important information used when
 * the CPU starts up. This important information includes the default clock ratio and
 * FSB speed. Strangely, the clock ratio tables used by the EBL_CR_POWERON ratio
 * are slightly different than the longhaul clock ratio tables, so we need those tables also.
 */

static int __initdata samuel2_EBL_CR_table[17] = {
	50,  /* 0000 ->  5.0x */
	30,  /* 0001 ->  3.0x */
	40,  /* 0010 ->  4.0x */
	100, /* 0011 -> 10.0x */
	55,  /* 0100 ->  5.5x */
	35,  /* 0101 ->  3.5x */
	45,  /* 0110 ->  4.5x */
	95,  /* 0111 ->  9.5x */
	90,  /* 1000 ->  9.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	60,  /* 1011 ->  6.0x */
	120, /* 1100 -> 12.0x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	65,  /* 1111 ->  6.5x */
  -2,  /* end of table  */
};

static int __initdata ezra_EBL_CR_table[16] = {
	50,  /* 0000 ->  5.0x */
	30,  /* 0001 ->  3.0x */
	40,  /* 0010 ->  4.0x */
	100, /* 0011 -> 10.0x */
	55,  /* 0100 ->  5.5x */
	35,  /* 0101 ->  3.5x */
	45,  /* 0110 ->  4.5x */
	95,  /* 0111 ->  9.5x */
	90,  /* 1000 ->  9.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	60,  /* 1011 ->  6.0x */
	120, /* 1100 -> 12.0x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	65,  /* 1111 ->  6.5x */
};

static int __initdata ezra_t_EBL_CR_table[33] = {
	50,  /* 0000 ->  5.0x */
	30,  /* 0001 ->  3.0x */
	40,  /* 0010 ->  4.0x */
	100, /* 0011 -> 10.0x */
	55,  /* 0100 ->  5.5x */
	35,  /* 0101 ->  3.5x */
	45,  /* 0110 ->  4.5x */
	95,  /* 0111 ->  9.5x */
	90,  /* 1000 ->  9.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	60,  /* 1011 ->  6.0x */
	120, /* 1100 -> 12.0x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	65,  /* 1111 ->  6.5x */
	-1,  /* 0000 -> RESERVED (9.0x) */
	110, /* 0001 -> 11.0x */
	120, /* 0010 -> 12.0x */
	-1,  /* 0011 -> RESERVED (10.0x)*/
	135, /* 0100 -> 13.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	105, /* 0111 -> 10.5x */
	130, /* 1000 -> 13.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	140, /* 1011 -> 14.0x */
	-1,  /* 1100 -> RESERVED (12.0x) */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	145, /* 1111 -> 14.5x */
  -2   /* end of table  */
};

static int __initdata Nehemiah_a_EBL_CR_table[33] = {
	50,  /* 0000 ->  5.0x */
	160, /* 0001 -> 16.0x */
	-1,  /* 0010 ->  RESERVED */
	100, /* 0011 -> 10.0x */
	55,  /* 0100 ->  5.5x */
	-1,  /* 0101 ->  RESERVED */
	-1,  /* 0110 ->  RESERVED */
	95,  /* 0111 ->  9.5x */
	90,  /* 1000 ->  9.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	60,  /* 1011 ->  6.0x */
	120, /* 1100 -> 12.0x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	65,  /* 1111 ->  6.5x */
	90,  /* 0000 ->  9.0x */
	-1,  /* 0001 -> RESERVED */
	120, /* 0010 -> 12.0x */
	100, /* 0011 -> 10.0x */
	135, /* 0100 -> 13.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	105, /* 0111 -> 10.5x */
	130, /* 1000 -> 13.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	140, /* 1011 -> 14.0x */
	120, /* 1100 -> 12.0x */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	145, /* 1111 -> 14.5x */
  -2   /* end of table  */
};
static int __initdata Nehemiah_b_EBL_CR_table[33] = {
	50,  /* 0000 ->  5.0x */
	160, /* 0001 -> 16.0x */
	-1,  /* 0010 ->  RESERVED */
	100, /* 0011 -> 10.0x */
	55,  /* 0100 ->  5.5x */
	-1,  /* 0101 ->  RESERVED */
	-1,  /* 0110 ->  RESERVED */
	95,  /* 0111 ->  9.5x */
	90,  /* 1000 ->  9.0x */
	70,  /* 1001 ->  7.0x */
	80,  /* 1010 ->  8.0x */
	60,  /* 1011 ->  6.0x */
	120, /* 1100 -> 12.0x */
	75,  /* 1101 ->  7.5x */
	85,  /* 1110 ->  8.5x */
	65,  /* 1111 ->  6.5x */
	90,  /* 0000 ->  9.0x */
	110, /* 0001 -> 11.0x */
	120, /* 0010 -> 12.0x */
	100, /* 0011 -> 10.0x */
	135, /* 0100 -> 13.5x */
	115, /* 0101 -> 11.5x */
	125, /* 0110 -> 12.5x */
	105, /* 0111 -> 10.5x */
	130, /* 1000 -> 13.0x */
	150, /* 1001 -> 15.0x */
	160, /* 1010 -> 16.0x */
	140, /* 1011 -> 14.0x */
	120, /* 1100 -> 12.0x */
	155, /* 1101 -> 15.5x */
	-1,  /* 1110 -> RESERVED (13.0x) */
	145, /* 1111 -> 14.5x */
  -2   /* end of table  */
};


/*
 * These are my 64 bit versions of wrmsr and rdmsr.
 * Some people prefer macros.
 *
 */

static void my_wrmsrl(int msr, unsigned long long value)
{
	unsigned long long long64_lo, long64_hi;

  long64_lo = value & 0xffffffff ;
  long64_hi = value >>32;
  wrmsr(msr, (__u32) long64_lo, (__u32) long64_hi);
}

static unsigned long long my_rdmsrl(int msr)
{
	unsigned long lo, hi;
  unsigned long long value;
  
  rdmsr (msr, lo, hi);
  value = hi;
  value = lo | (value<<32);
  return value;
}

/*
 *
 *  load_longhaul_state  and   save_longhaul_state
 *
 *    I don't like globals, so I create a longhaul_state object in the module_init function
 *    and fill it with data. Problem is, there are four entry points to this module: module_init,
 *    module_cleanup, and the two callbacks. To transport the longhaul_state object between
 *    these entry points I need a persistent storage buffer (which is a happens to be a global
 *    variable - yuck!). And I use the save_longhaul_state and load_loadhaul_state functions
 *    to copy the obejct to and from the persistent storage area.
 *
 *    This is a yucky OO hack, but not as bad as GTK+. (haha)
 *
 */

void load_longhaul_state(struct longhaul_state_t *longhaul_state)
{
	memcpy (longhaul_state, &longhaul_state_buffer, sizeof(struct longhaul_state_t));
}

static void save_longhaul_state(struct longhaul_state_t *longhaul_state)
{
	memcpy (&longhaul_state_buffer, longhaul_state, sizeof(struct longhaul_state_t));
}

static int longhaul_get_cpu_fsb (struct longhaul_state_t *longhaul_state)
{
	unsigned int eblcr_fsb_table[] = { 66, 133, 100, -1 };
	unsigned long invalue=0,lo, hi;

	rdmsr (MSR_IA32_EBL_CR_POWERON, lo, hi);
	invalue = (lo & (1<<18|1<<19)) >>18;
	return eblcr_fsb_table[invalue];
}

static int longhaul_get_cpu_mult (struct longhaul_state_t *longhaul_state)
{
	unsigned long invalue=0,lo, hi;

  if (longhaul_state->version == 0) return 0;

	rdmsr (MSR_IA32_EBL_CR_POWERON, lo, hi);
	invalue = (lo & (1<<22|1<<23|1<<24|1<<25)) >>22;
	if ((longhaul_state->version == 3) && (lo & (1<<27))) invalue+=16;
	if ((longhaul_state->version == 4) && (lo & (1<<27))) invalue+=16;
	
	return longhaul_state->EBL_CR_table[invalue];
}

/*
 * A couple simple functions to get the version,
 *
 */

static int longhaul_get_version (struct longhaul_state_t *longhaul_state)
{
	struct cpuinfo_x86 *c = cpu_data;
	int version;

  longhaul_state->can_scale_voltage = 0;
  longhaul_state->cpu_family = c->x86;
  longhaul_state->cpu_model = c->x86_model;
  longhaul_state->cpu_stepping = c->x86_mask;
  
	version = 0;
	switch (longhaul_state->cpu_model) {
	case 7:    /* Samuel II and Ezra */
		switch (longhaul_state->cpu_stepping) {
		case 1 ... 7:		/* Samuel II */
			version=2;
      longhaul_state->longhaul_cpu_mult_table = longhaul2_cpu_mult_table;
      longhaul_state->EBL_CR_table = samuel2_EBL_CR_table;
      break;
    case 8 ... 15:  /* Ezra  */
      version=2;
      longhaul_state->longhaul_cpu_mult_table = longhaul2_cpu_mult_table;
      longhaul_state->EBL_CR_table = ezra_EBL_CR_table;      
			break;
		}
		break;

	case 8:		/* Ezra-T   */
    version=3;
    longhaul_state->longhaul_cpu_mult_table = longhaul3_cpu_mult_table;
    longhaul_state->EBL_CR_table = ezra_t_EBL_CR_table;      
    break;

  case 9:   /* Nehemiah */
    switch (longhaul_state->cpu_stepping) {
    case 0 ... 1:
      version=4;
      longhaul_state->longhaul_cpu_mult_table = longhaul4a_cpu_mult_table;
      longhaul_state->EBL_CR_table = Nehemiah_a_EBL_CR_table;      
      break;
    case 2 ... 15:
      version=4;
      longhaul_state->longhaul_cpu_mult_table = longhaul4b_cpu_mult_table;
      longhaul_state->EBL_CR_table = Nehemiah_b_EBL_CR_table;      
      break;
    }  
  }
  
	return version;
}

/*
 * sort the cpu multiplier list from lowest to highest
 */

static void  longhaul_sort_clock_table(struct clock_table_t *clock_table)
{
  int last, changed=1, size, i;
  void *swap_buffer;
  
  for (last=0; clock_table[last].frequency != CLOCK_TABLE_END; last++);

  if (last == 1) return;  /* nothing to sort - only one entry */
  
  size = sizeof(struct clock_table_t);
  swap_buffer = kmalloc(size, GFP_KERNEL);

  while (changed == 1) {      /* do until a clean run with no changes mean all sorted */
    changed = 0;
    for (i=1; i<last; i++) {
      if (clock_table[i-1].frequency > clock_table[i].frequency) {
        memcpy(swap_buffer, &clock_table[i-1], size);
        memcpy(&clock_table[i-1],&clock_table[i], size);
        memcpy(&clock_table[i], swap_buffer, size);
        changed = 1;
      }
    }
  }
  kfree(swap_buffer);
}

static int longhaul_get_clock_table (struct longhaul_state_t *longhaul_state)
{
	unsigned long invalue;
	unsigned int j, k, minmult=0, maxmult=0;
	unsigned int multipliers[32]= {
		50,30,40,100,55,35,45,95,90,70,80,60,120,75,85,65,
		-1,110,120,-1,135,115,125,105,130,150,160,140,-1,155,-1,145 };
	union msr_longhaul longhaul;
  struct clock_table_t *clock_table;
  int *longhaul_cpu_mult_table;

  if (longhaul_state->version == 0) return 0;

	longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
	
	invalue = longhaul.bits.MaxMHzBR;
	if (longhaul.bits.MaxMHzBR4) invalue += 16;
	maxmult=multipliers[invalue];
  dprintk(KERN_INFO PFX" invalue: %ld  maxmult: %d \n", invalue, maxmult);

  /* In theory the minimum multiplier should be specified in MSR_VIA_LONGHAUL
   * with the MinMHZBR and MinMHZBR4 bits. But in reality that doesn't
   * seem to work. The following switch statement has been developed by trial
   * and error.
   *
   * So far tobbe has reported that his 6.9.1 freezes up at 3.0 and 3.5.
   */
  
  switch (longhaul_state->cpu_model) {
  case 9:
    minmult=40;
    break;
  default:
    minmult=30;
    break;
  }
	
	longhaul_state->highest_speed = maxmult * longhaul_state->fsb * 100;
	longhaul_state->lowest_speed  = minmult * longhaul_state->fsb * 100;

	clock_table = kmalloc(33*sizeof(struct clock_table_t), GFP_KERNEL);
	if(!clock_table) return -ENOMEM;
  
  longhaul_state->clock_table = clock_table;
  longhaul_cpu_mult_table = longhaul_state->longhaul_cpu_mult_table;
  
	for (k=0, j=0; longhaul_cpu_mult_table[j] != -2; j++) {
		if (longhaul_cpu_mult_table[j] == -1) continue;
    if (longhaul_cpu_mult_table[j] < minmult) continue;
    if (!OVERCLOCK && longhaul_cpu_mult_table[j] > maxmult) continue;
		clock_table[k].frequency= longhaul_cpu_mult_table[j] * longhaul_state->fsb * 100;
		clock_table[k].index	= (j << 8);
		k++;
	}
	
	for (j=k; j<=33; j++) clock_table[k].frequency = CLOCK_TABLE_END;
  
	if (!k) {
		kfree (clock_table);
		return -EINVAL;
	}

  longhaul_sort_clock_table(clock_table);
  
	return 0;
}

static int longhaul_get_status(struct longhaul_state_t *plonghaul_state)
{
	plonghaul_state->fsb = longhaul_get_cpu_fsb(plonghaul_state);
	plonghaul_state->version = longhaul_get_version(plonghaul_state);
	plonghaul_state->cpu_mult = longhaul_get_cpu_mult(plonghaul_state);
  plonghaul_state->cpu_speed = plonghaul_state->fsb * plonghaul_state->cpu_mult * 100;
	longhaul_get_clock_table(plonghaul_state);

	return 0;
}

/*
 *
 *    longhaul_set_state:  Sets new cpu multiplier
 *
 */

void longhaul_set_state (struct longhaul_state_t *longhaul_state,
                         unsigned int desired_clock_index)
{
	int cpu_mult_index;
	union msr_longhaul longhaul;
  struct clock_table_t *clock_table;

  if (longhaul_state->version == 0) return;

  clock_table = longhaul_state->clock_table;
  
	if (clock_table[desired_clock_index].frequency == CLOCK_TABLE_END) return;

	cpu_mult_index = clock_table[desired_clock_index].index >> 8;

	switch (longhaul_state->version) {
	case 1:
		printk (PFX " not implemented for longhaul 1 \n");
		break;
		
	case 2:
		longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
		dprintk (PFX " original         longhaul.val %llx\n", longhaul.val);
		
		longhaul.bits.SoftBusRatio = cpu_mult_index & 0xf;
		longhaul.bits.SoftBusRatio4 = (cpu_mult_index & 0x10) >> 4;
		longhaul.bits.EnableSoftBusRatio = 1;

    /* in the original version of the longhaul driver
     * the revision key was set to 3, however it seems to work better
     * if we set if to 0x0 here and 0xf later. I don't know why.
     */
		longhaul.bits.RevisionKey = 0x0;

		dprintk (PFX " before 1st write longhaul.val %llx\n", longhaul.val);
		my_wrmsrl(MSR_VIA_LONGHAUL, longhaul.val);
		__hlt();

		longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
		dprintk (PFX " after  1st write longhaul.val %llx\n", longhaul.val);

		longhaul.bits.EnableSoftBusRatio = 0;
		if (longhaul_state->can_scale_voltage) longhaul.bits.EnableSoftVID = 0;
		longhaul.bits.RevisionKey = 0x0f;

		dprintk (PFX " before 2nd write longhaul.val %llx\n", longhaul.val);
		my_wrmsrl(MSR_VIA_LONGHAUL, longhaul.val);

		longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
		dprintk (PFX " after  2nd write longhaul.val %llx\n", longhaul.val);
		
		break;
 
	case 3:
  case 4:
		longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
		dprintk (PFX " original         longhaul.val %llx\n",	longhaul.val);
		longhaul.bits.SoftBusRatio = cpu_mult_index & 0xf;
		longhaul.bits.SoftBusRatio4 = (cpu_mult_index & 0x10) >> 4;
		longhaul.bits.EnableSoftBusRatio = 1;
    
    longhaul.bits.RevisionKey = 0x0; 

		dprintk (PFX " before 2nd write longhaul.val %llx \n", longhaul.val);
		my_wrmsrl(MSR_VIA_LONGHAUL, longhaul.val);
		__hlt();

		longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
		longhaul.bits.EnableSoftBusRatio = 0;
		longhaul.bits.RevisionKey = 0x0f;
		my_wrmsrl(MSR_VIA_LONGHAUL, longhaul.val);
		break;
	}
	longhaul_state->cpu_mult = longhaul_get_cpu_mult(longhaul_state);
  longhaul_state->cpu_speed = longhaul_state->fsb * longhaul_state->cpu_mult * 100;
  save_longhaul_state(longhaul_state);
}
  

/*
 *
 *   longhaul_proc_read:  this is the callback that provides data when a user
 *                        reads the /proc/longhaul file
 *
 */


static int longhaul_proc_read(char *buf, char **start, off_t offset,
                              int count, int *eof, void *data)
{
	int i;
	char* p = buf;
	struct longhaul_state_t longhaul_state;
  struct cpuinfo_x86 *c = cpu_data;

	load_longhaul_state(&longhaul_state);     /* load the saved state     */
	longhaul_get_status(&longhaul_state);     /* refresh the state struct */

  switch (longhaul_state.version) {
  case 2:	/* version 2 is supported */
  case 3: /* version 3 is supported */
  case 4: /* version 4 is supported */
		

    p += sprintf (p, "CPU %d.%d.%d \n", c->x86, c->x86_model, c->x86_mask);
    p += sprintf (p, "CPU Speed: %ld\n", longhaul_state.cpu_speed);
		p += sprintf (p, "FSB: %3d\n", longhaul_state.fsb);
		p += sprintf (p, "CPU Mult: %3d\n", longhaul_state.cpu_mult);
		p += sprintf (p, "Longhaul version: %1x\n", longhaul_state.version);
		p += sprintf (p, "Driver version: %1d.%d\n", DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION);
		p += sprintf (p, "last command: %x\n", longhaul_state.last_command);
    p += sprintf (p, "                            7 6 5 4 3 2 1 0\n");
    p += sprintf (p, "MSR_VIA_LONGHAUL:        %#018llx \n", my_rdmsrl (MSR_VIA_LONGHAUL));
    p += sprintf (p, "MSR_IA32_EBL_CR_POWERON: %#018llx \n", my_rdmsrl (MSR_IA32_EBL_CR_POWERON));

		for (i=0; longhaul_state.clock_table[i].frequency != CLOCK_TABLE_END; i++)
			p += sprintf (p, "clock %#x  %06d\n", i, longhaul_state.clock_table[i].frequency);	

    break;
    
  default:
    		
	/* other versions are not supported */

		p += sprintf (p, "Your processor is not supported.\n");
		p += sprintf (p, "Please send the following info to andrew@howlett.net:\n\n");
    p += sprintf (p, "CPU: %d.%d.%d \n", c->x86, c->x86_model, c->x86_mask);        
    p += sprintf (p, "CPU Speed: %ld\n", longhaul_state.cpu_speed);
		p += sprintf (p, "FSB: %3d\n", longhaul_state.fsb);
		p += sprintf (p, "CPU Mult: %3d\n", longhaul_state.cpu_mult);
		p += sprintf (p, "Longhaul version: %1x\n", longhaul_state.version);
		p += sprintf (p, "Driver version: %1d.%d\n", DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION);
		p += sprintf (p, "last command: %x\n", longhaul_state.last_command);
    p += sprintf (p, "                            7 6 5 4 3 2 1 0\n");
    p += sprintf (p, "MSR_VIA_LONGHAUL:        %#018llx \n", my_rdmsrl (MSR_VIA_LONGHAUL));
    p += sprintf (p, "MSR_IA32_EBL_CR_POWERON: %#018llx \n", my_rdmsrl (MSR_IA32_EBL_CR_POWERON));
    break;    
	}
	
	*eof=1;
	return (p - buf);
}

/*
 *
 *     longhaul_proc_write:  This is the callback which accepts data when a user
 *                           writes to the /proc/longhaul file
 *
 */

static int longhaul_proc_write(struct file *file, const char *buffer, unsigned long count,
                               void *data)
{
	struct longhaul_state_t longhaul_state;
  unsigned int input = ~0;
  
	load_longhaul_state(&longhaul_state);     /* load the saved state     */
	longhaul_get_status(&longhaul_state);     /* refresh the state object */
  
  if (count == 0) return 0;
  
  switch (buffer[0]) {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    input = buffer[0] - 0x30;
    break;
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
    input = buffer[0] - 0x61 + 10;
    break;
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
    input = buffer[0] - 0x41 + 10;
    break;
  }
  if (input == ~0) return count;  /* bogus command */

  longhaul_state.last_command = input;
  save_longhaul_state(&longhaul_state);

  if (longhaul_state.version == 0) return count;   /* version of longhaul not supported */
    
  if (longhaul_state.clock_table[input].frequency == CLOCK_TABLE_END) return count; /* invalid */
  
  longhaul_set_state(&longhaul_state, input);
  
  return count;
}

static int longhaul_set_bus_speed(struct longhaul_state_t *longhaul_state,
                                  unsigned int new_speed)
{
	union msr_longhaul longhaul;

  if (longhaul_state->version == 0 ) return 0;
  
  
  longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
  dprintk(KERN_INFO PFX " original         longhaul.val %llx \n", longhaul.val);

  //  longhaul.bits.RevisionID = 0xf;
  
  longhaul.bits.SoftBSEL = new_speed; 
  longhaul.bits.EnableSoftBSEL = 1; 
/*   /\* We must program the revision key only with values we */
/*    * know about, not blindly copy it from 0:3 *\/ */
  longhaul.bits.RevisionKey = 0x0f; /* have to use revision key 0x00 */
  
  dprintk(KERN_INFO PFX " before 1st write longhaul.val %llx \n", longhaul.val);
  my_wrmsrl(MSR_VIA_LONGHAUL, longhaul.val); 
  __hlt(); 

  longhaul.val = my_rdmsrl (MSR_VIA_LONGHAUL);
  dprintk(KERN_INFO PFX " after  1st write longhaul.val %llx \n", longhaul.val);
  
  return 0;
}


/*
 *
 *     init_module:  When the module is installed it executes this function
 *
 */

int init_module(void) 
{
	struct longhaul_state_t longhaul_state;

	/* create the /proc/longhaul file */
  
	longhaul_state.proc_file = create_proc_entry("longhaul",         /* proc file name */
                                                0644,              /* default mode   */
                                                NULL);             /* parent dir     */ 
  if (longhaul_state.proc_file == NULL) {
    printk(KERN_INFO PFX " failed to create /proc/longhaul\n");
    return -ENOMEM;
  }

  longhaul_state.proc_file->read_proc = longhaul_proc_read;
  longhaul_state.proc_file->write_proc = longhaul_proc_write;
  longhaul_state.proc_file->owner = THIS_MODULE;

  /* create longhaul_status object (I use object very loosely) */
  
  longhaul_state.last_command = ~0;
  longhaul_state.overclock_on = OVERCLOCK;
 	longhaul_get_status(&longhaul_state);

  //  longhaul_set_bus_speed(&longhaul_state, 1);
  
  //	longhaul_set_state(&longhaul_state, 3);

	longhaul_get_status(&longhaul_state);
	save_longhaul_state(&longhaul_state);       /* make longhaul_state object available */

  printk(KERN_INFO PFX " installed \n");
  
	return 0;                                   /* to callbacks and cleanup             */
}

void cleanup_module(void)
{
	struct longhaul_state_t longhaul_state;
	
	remove_proc_entry("longhaul", NULL);

	load_longhaul_state(&longhaul_state);   /* need the clock_table pointer */
	kfree (longhaul_state.clock_table);
	printk(KERN_INFO PFX " removed\n");
}


MODULE_AUTHOR ("adapted by A. Howlett <andrew@howlett.net> \nfrom code by Dave Jones <davej@codemonkey.org.uk>");
MODULE_DESCRIPTION ("Longhaul driver for VIA C3 processors.");
MODULE_LICENSE ("GPL");

//EXPORT_SYMTAB;

//EXPORT_SYMBOL(longhaul_set_state);
//EXPORT_SYMBOL(load_longhaul_state);


/*
 * Overrides for Emacs so that we get a uniform tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 2
 * End:
 */
